import { __awaiter, __decorate } from "tslib";
import { Auth, Connect, logger, } from '@genesislcap/foundation-comms';
import { UUID } from '@genesislcap/foundation-utils';
import { getAvailableIndexes, getAvailableIndexFields } from '../utils';
export class BaseServerSideDatasource {
    constructor(options) {
        this.rowData = new Map();
        /**
         * The number of rows in the grid. THIS DO NOT CONSIDER CRITERIA_MATCH scenarios.
         * @internal
         */
        this.serverRowsCount = 0;
        /**
         * The number of rows in the grid. THIS IS A LOCAL COUNT CONSIDERING CONSIDER CRITERIA_MATCH scenarios.
         * @internal
         */
        this.clientRowsCount = 0;
        this.moreRows = false;
        this.calculatedRowsCount = 0;
        this.currentSequenceId = null;
        this.reloadResourceDataFunc = options.reloadResourceDataFunc;
        this.resourceName = options.resourceName;
        this.resourceParams = options.resourceParams;
        this.originalCriteriaMatch = options.resourceParams.CRITERIA_MATCH;
        this.resourceIndexes = options.resourceIndexes;
        this.resourceColDefs = options.resourceColDefs;
        this.maxRows = options.maxRows;
        this.maxView = options.maxView;
        this.rowId = options.rowId;
        this.pagination = options.pagination;
        this.zeroBasedViewNumber = options.zeroBasedViewNumber;
    }
    getRows(params) {
        throw new Error('Method not implemented.');
    }
    setupFilteringAndSorting(params) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const filterModelBeingApplied = params.request.filterModel;
            const filtersAreBeingApplied = Object.keys(filterModelBeingApplied).length > 0;
            const filtersAreGettingCleaned = Object.keys(filterModelBeingApplied).length === 0;
            const filterModelBeingAppliedDiffersFromCurrent = JSON.stringify(filterModelBeingApplied) !== JSON.stringify(this.currentFilterModel);
            if (filterModelBeingAppliedDiffersFromCurrent || filtersAreBeingApplied) {
                if (filtersAreGettingCleaned && this.currentFilterModel) {
                    this.currentFilterModel = null;
                    this.resourceParams.CRITERIA_MATCH = this.originalCriteriaMatch;
                    yield this.refreshDatasource(params);
                    return;
                }
                else if (filterModelBeingAppliedDiffersFromCurrent && filtersAreBeingApplied) {
                    this.currentFilterModel = filterModelBeingApplied;
                    this.resourceParams.CRITERIA_MATCH = this.buildCriteriaMatchFromFilters();
                    yield this.refreshDatasource(params);
                    return;
                }
            }
            const toBeAppliedSortModel = params.request.sortModel;
            if (((_a = this.currentSortModel) === null || _a === void 0 ? void 0 : _a.length) !== toBeAppliedSortModel.length ||
                toBeAppliedSortModel.length > 0) {
                if (toBeAppliedSortModel.length === 0 && this.currentSortModel) {
                    this.calculatedRowsCount = 0;
                    this.currentSortModel = null;
                    this.resourceParams.ORDER_BY = null;
                    this.resourceParams.REVERSE = null;
                    yield this.refreshDatasource(params);
                    return;
                }
                else if (toBeAppliedSortModel.length > 0) {
                    this.currentSortModel = toBeAppliedSortModel;
                    const coldIdBeingSorted = toBeAppliedSortModel[0].colId; // Not allowing multiple sorts by user
                    const sortTypeBeingApplied = toBeAppliedSortModel[0].sort;
                    const orderByAndToBeSortedColIds = this.getOrderByAndToBeSortedColIds(this.resourceIndexes, coldIdBeingSorted);
                    if (!orderByAndToBeSortedColIds) {
                        this.calculatedRowsCount = 0;
                        const availableIndexes = getAvailableIndexes(this.resourceIndexes);
                        const availableIndexFields = getAvailableIndexFields(this.resourceIndexes);
                        logger.warn('The FIELD/column (${coldIdBeingSorted}) being sorted is not part of an INDEX, required for the [orderBy] operation. See https://learn.genesis.global/docs/database/data-types/index-entities/');
                        logger.debug('Available indexes:', availableIndexes);
                        logger.debug('Columns that can be sorted with the available indexes:', availableIndexFields);
                        params.fail();
                        return;
                    }
                    else if (JSON.stringify(toBeAppliedSortModel) !== JSON.stringify(this.currentSortModel)) {
                        this.resourceParams.ORDER_BY = orderByAndToBeSortedColIds.orderBy;
                        this.resourceParams.REVERSE = sortTypeBeingApplied === 'desc' ? true : false;
                        yield this.refreshDatasource(params);
                        return;
                    }
                }
            }
        });
    }
    destroy() {
        this.rowData = new Map();
        this.clientRowsCount = 0;
        this.calculatedRowsCount = 0;
        this.serverRowsCount = 0;
        this.currentSequenceId = 0;
        this.moreRows = false;
        this.sourceRef = undefined;
        this.lastSuccessRowData = undefined;
    }
    refreshDatasource(params) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.destroy();
            yield this.reloadResourceDataFunc({
                criteria: this.resourceParams.CRITERIA_MATCH,
                orderBy: this.resourceParams.ORDER_BY,
                reverse: this.resourceParams.REVERSE,
            });
        });
    }
    getOrderByAndToBeSortedColIds(map, searchValue) {
        for (const [key, value] of map.entries()) {
            const searchIndex = value.findIndex((x) => x === searchValue);
            if (searchIndex !== -1) {
                const copy = [...value];
                const toBeSorted = copy.splice(searchIndex, 1);
                return { orderBy: key, toBeSorted };
            }
        }
    }
    buildCriteriaMatchFromFilters() {
        this.resourceParams.CRITERIA_MATCH = this.originalCriteriaMatch;
        if (this.resourceParams.CRITERIA_MATCH &&
            typeof this.resourceParams.CRITERIA_MATCH === 'string') {
            this.resourceParams.CRITERIA_MATCH = this.resourceParams.CRITERIA_MATCH.split(';');
        }
        if (!this.resourceParams.CRITERIA_MATCH) {
            this.resourceParams.CRITERIA_MATCH = [];
        }
        const criteria = this.resourceParams.CRITERIA_MATCH.concat(this.criteriaFromFilters());
        return criteria.join(' && ');
    }
    criteriaFromFilters() {
        const filters = [];
        this.getFiltersByType('text').forEach((k) => {
            if (!this.currentFilterModel[k].filter &&
                (this.currentFilterModel[k].type === 'false' || this.currentFilterModel[k].type === 'true')) {
                filters.push(`${k} == ${this.currentFilterModel[k].type}`);
            }
            else {
                switch (this.currentFilterModel[k].type) {
                    case 'blank':
                        filters.push(`${k} == ''`);
                        break;
                    case 'contains':
                        filters.push(`Expr.containsIgnoreCase(${k}, '${this.currentFilterModel[k].filter}')`);
                        break;
                    case 'equals':
                        filters.push(`${k} == '${this.currentFilterModel[k].filter}'`);
                        break;
                    case 'notBlank':
                        filters.push(`${k} != ''`);
                        break;
                    case 'notEqual':
                        filters.push(`${k} != '${this.currentFilterModel[k].filter}'`);
                        break;
                    case 'wordStartsWith':
                        filters.push(`Expr.containsWordsStartingWithIgnoreCase(${k}, '${this.currentFilterModel[k].filter}')`);
                        break;
                }
            }
        });
        this.getFiltersByType('number').forEach((k) => {
            const value = this.currentFilterModel[k].filter;
            const valueTwo = this.currentFilterModel[k].filterTo;
            switch (this.currentFilterModel[k].type) {
                case 'equals':
                    !isNaN(value) && filters.push(`${k} == ${value}`);
                    break;
                case 'notEqual':
                    !isNaN(value) && filters.push(`${k} != ${value}`);
                    break;
                case 'greaterThan':
                    !isNaN(value) && filters.push(`${k} > ${value}`);
                    break;
                case 'greaterThanOrEqual':
                    !isNaN(value) && filters.push(`${k} >= ${value}`);
                    break;
                case 'lessThan':
                    !isNaN(value) && filters.push(`${k} < ${value}`);
                    break;
                case 'lessThanOrEqual':
                    !isNaN(value) && filters.push(`${k} <= ${value}`);
                    break;
                case 'inRange':
                    !isNaN(value) &&
                        !isNaN(valueTwo) &&
                        filters.push(`${k} >= ${value} && ${k} <= ${valueTwo}`);
                    break;
                case 'blank':
                    filters.push(`${k} == 0`);
                    break;
                case 'notBlank':
                    filters.push(`${k} != 0`);
                    break;
            }
        });
        this.getFiltersByType('date').forEach((k) => {
            var _a, _b;
            const dateFrom = (_a = this.currentFilterModel[k].dateFrom) === null || _a === void 0 ? void 0 : _a.replace(/-/g, '').replace('T', '-').split(' ')[0];
            const dateTo = (_b = this.currentFilterModel[k].dateTo) === null || _b === void 0 ? void 0 : _b.replace(/-/g, '').replace('T', '-').split(' ')[0];
            switch (this.currentFilterModel[k].type) {
                case 'equals':
                    filters.push(`Expr.dateIsEqual(${k}, '${dateFrom}')`);
                    break;
                case 'lessThan':
                    filters.push(`Expr.dateIsBefore(${k}, '${dateFrom}')`);
                    break;
                case 'greaterThan':
                    filters.push(`Expr.dateIsAfter(${k}, '${dateFrom}')`);
                    break;
                case 'inRange':
                    filters.push(`Expr.dateIsAfter(${k}, '${dateFrom}') && Expr.dateIsBefore(${k}, '${dateTo}')`);
                    break;
                case 'isToday':
                    const now = new Date();
                    const year = now.getFullYear();
                    const month = (now.getMonth() + 1).toString().padStart(2, '0');
                    const day = now.getDate().toString().padStart(2, '0');
                    const todayStr = `${year}-${month}-${day} 00:00:00`;
                    filters.push(`Expr.dateIsEqual(${k}, '${todayStr}')`);
                    break;
                case 'blank':
                    filters.push(`${k} == null`);
                    break;
                case 'notBlank':
                    filters.push(`${k} != null`);
                    break;
            }
        });
        return filters;
    }
    getFiltersByType(filterType) {
        return Object.keys(this.currentFilterModel)
            .filter((k) => this.resourceColDefs.findIndex((o) => o.NAME === k) !== -1)
            .filter((k) => this.currentFilterModel[k].filterType === filterType);
    }
    getPageNumber(start, recordsPerPage, totalRecords, zeroBasedViewNumber = true) {
        if (start > totalRecords) {
            start = Math.floor((totalRecords - 1) / recordsPerPage) * recordsPerPage + 1;
        }
        const currentPageIndex = Math.ceil(start / recordsPerPage);
        return zeroBasedViewNumber ? currentPageIndex : currentPageIndex + 1;
    }
    getCorrectRowCount(params) {
        const currentLastRowNumber = params.request.endRow;
        this.calculatedRowsCount = this.clientRowsCount || currentLastRowNumber;
        const defaultCount = Math.min(this.serverRowsCount, this.maxView);
        let rowCount;
        if (this.pagination) {
            rowCount = !this.resourceParams.CRITERIA_MATCH
                ? Math.min(this.serverRowsCount, this.maxView)
                : undefined;
            if (this.resourceParams.CRITERIA_MATCH) {
                rowCount = defaultCount;
            }
            return rowCount;
        }
        if (currentLastRowNumber === this.serverRowsCount) {
            rowCount = this.serverRowsCount;
        }
        else if (currentLastRowNumber > this.maxView ||
            currentLastRowNumber > this.calculatedRowsCount) {
            rowCount = Math.min(defaultCount, this.calculatedRowsCount);
        }
        else if (!this.moreRows && this.serverRowsCount > this.calculatedRowsCount) {
            rowCount = this.calculatedRowsCount;
        }
        return rowCount;
    }
    get isNotFirstViewNumber() {
        return this.zeroBasedViewNumber ? this.currentSequenceId > 0 : this.currentSequenceId > 1;
    }
}
__decorate([
    Auth
], BaseServerSideDatasource.prototype, "auth", void 0);
__decorate([
    Connect
], BaseServerSideDatasource.prototype, "connect", void 0);
__decorate([
    UUID
], BaseServerSideDatasource.prototype, "uuid", void 0);
